@@ -102,16 +102,16 @@ static void vma_free (VMA_t *vma)
   g_free (vma->private_clean);
   g_free (vma->private_dirty);
 
-  g_slice_free (VMA_t, vma);
+  g_free (vma);
 }
 
 static void perm_entry_free (PermEntry *entry)
 {
-  if (entry)
-    g_slice_free (PermEntry, entry);
+  g_free (entry);
 }
 
-static void add_to_perm_entry (GHashTable *hash, VMA_t *entry)
+static void add_to_perm_entry (GHashTable *hash,
+                               VMA_t      *entry)
 {
   const char *perms = entry->perms;
   PermEntry *value;
@@ -121,7 +121,7 @@ static void add_to_perm_entry (GHashTable *hash, VMA_t *entry)
   value = g_hash_table_lookup (hash, perms);
 
   if (!value) {
-    value = g_slice_new0 (PermEntry);
+    value = g_new0 (PermEntry, 1);
     insert = TRUE;
   }
 
@@ -138,7 +138,8 @@ static void add_to_perm_entry (GHashTable *hash, VMA_t *entry)
     g_hash_table_insert (hash, g_strdup (perms), value);
 }
 
-static void add_to_totals (PermEntry *entry, PermEntry *totals)
+static void add_to_totals (PermEntry *entry,
+                           PermEntry *totals)
 {
   totals->shared_clean += entry->shared_clean;
   totals->shared_dirty += entry->shared_dirty;
@@ -146,7 +147,9 @@ static void add_to_totals (PermEntry *entry, PermEntry *totals)
   totals->private_dirty += entry->private_dirty;
 }
 
-static void print_vma_table (GString *str, GHashTable *hash, const char *caption)
+static void print_vma_table (GString    *str,
+                             GHashTable *hash,
+                             const char *caption)
 {
   PermEntry *pentry, totals;
 
@@ -189,7 +192,10 @@ static void print_vma_table (GString *str, GHashTable *hash, const char *caption
   g_string_append (str, "</table>");
 }
 
-static void ephy_smaps_pid_to_html (EphySMaps *smaps, GString *str, pid_t pid, EphyProcess process)
+static void ephy_smaps_pid_to_html (EphySMaps  *smaps,
+                                    GString    *str,
+                                    pid_t       pid,
+                                    EphyProcess process)
 {
   GFileInputStream *stream;
   GDataInputStream *data_stream;
@@ -228,7 +234,7 @@ static void ephy_smaps_pid_to_html (EphySMaps *smaps, GString *str, pid_t pid, E
       if (vma)
         vma_entries = g_slist_append (vma_entries, vma);
 
-      vma = g_slice_new0 (VMA_t);
+      vma = g_new0 (VMA_t, 1);
 
       vma->start = g_match_info_fetch (match_info, 1);
       vma->end = g_match_info_fetch (match_info, 2);
@@ -272,7 +278,7 @@ static void ephy_smaps_pid_to_html (EphySMaps *smaps, GString *str, pid_t pid, E
     }
 
     g_match_info_free (match_info);
- out:
+out:
     g_free (line);
   }
 
@@ -408,7 +414,9 @@ static EphyProcess get_ephy_process (pid_t pid)
   return process;
 }
 
-static void ephy_smaps_pid_children_to_html (EphySMaps *smaps, GString *str, pid_t parent_pid)
+static void ephy_smaps_pid_children_to_html (EphySMaps *smaps,
+                                             GString   *str,
+                                             pid_t      parent_pid)
 {
   GDir *proc;
   const char *name;
