@@ -33,6 +33,7 @@
 #include "ephy-prefs.h"
 #include "ephy-settings.h"
 #include "ephy-shell.h"
+#include "ephy-tab-label.h"
 #include "ephy-window.h"
 
 #include <glib/gi18n.h>
@@ -78,8 +79,7 @@ static void ephy_notebook_page_reordered (GtkNotebook *notebook,
                                           GtkWidget   *child,
                                           guint        page_num);
 
-static const GtkTargetEntry url_drag_types [] =
-{
+static const GtkTargetEntry url_drag_types [] = {
   { (char *)"GTK_NOTEBOOK_TAB", GTK_TARGET_SAME_APP, 0 },
   { (char *)EPHY_DND_URI_LIST_TYPE, 0, 0 },
   { (char *)EPHY_DND_URL_TYPE, 0, 1 },
@@ -177,7 +177,9 @@ ephy_notebook_class_init (EphyNotebookClass *klass)
 }
 
 static gint
-find_tab_num_at_pos (EphyNotebook *notebook, gint abs_x, gint abs_y)
+find_tab_num_at_pos (EphyNotebook *notebook,
+                     gint          abs_x,
+                     gint          abs_y)
 {
   int page_num = 0;
   GtkNotebook *nb = GTK_NOTEBOOK (notebook);
@@ -224,8 +226,11 @@ button_press_cb (EphyNotebook   *notebook,
 
   if (event->type == GDK_BUTTON_PRESS && event->button == GDK_BUTTON_MIDDLE) {
     GtkWidget *tab = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), tab_clicked);
-    g_signal_emit (notebook, signals[TAB_CLOSE_REQUEST], 0, tab);
-    return GDK_EVENT_STOP;
+
+    if (!ephy_notebook_tab_is_pinned (notebook, EPHY_EMBED (tab))) {
+      g_signal_emit (notebook, signals[TAB_CLOSE_REQUEST], 0, tab);
+      return GDK_EVENT_STOP;
+    }
   }
 
   if (event->type == GDK_BUTTON_PRESS &&
@@ -254,9 +259,15 @@ ephy_notebook_switch_page_cb (GtkNotebook *notebook,
 {
   EphyNotebook *nb = EPHY_NOTEBOOK (notebook);
   GtkWidget *child;
+  GtkWidget *tab_label;
 
   child = gtk_notebook_get_nth_page (notebook, page_num);
-  gtk_widget_grab_focus (child);
+  if (!ephy_web_view_is_in_auth_dialog (ephy_embed_get_web_view (EPHY_EMBED (child))))
+    gtk_widget_grab_focus (child);
+
+  tab_label = gtk_notebook_get_tab_label (notebook, page);
+  if (ephy_tab_label_is_pinned (tab_label))
+    ephy_tab_label_set_needs_attention (tab_label, FALSE);
 
   /* Remove the old page, we dont want to grow unnecessarily
    * the list */
@@ -365,7 +376,7 @@ update_tabs_visibility (EphyNotebook *nb,
   if (mode != EPHY_EMBED_SHELL_MODE_APPLICATION &&
       nb->adaptive_mode != EPHY_ADAPTIVE_MODE_NARROW &&
       ((policy == EPHY_PREFS_UI_TABS_BAR_VISIBILITY_POLICY_MORE_THAN_ONE && num > 1) ||
-        policy == EPHY_PREFS_UI_TABS_BAR_VISIBILITY_POLICY_ALWAYS))
+       policy == EPHY_PREFS_UI_TABS_BAR_VISIBILITY_POLICY_ALWAYS))
     show_tabs = TRUE;
 
   /* Only show the tabs when the "tabs-allowed" property is TRUE. */
@@ -387,9 +398,10 @@ expand_tabs_changed_cb (GSettings    *settings,
   tabs = gtk_container_get_children (GTK_CONTAINER (nb));
 
   for (l = tabs; l != NULL; l = l->next) {
+    gboolean pinned = ephy_notebook_tab_is_pinned (nb, l->data);
     gtk_container_child_set (GTK_CONTAINER (nb),
                              l->data,
-                             "tab-expand", expand,
+                             "tab-expand", pinned ? FALSE : expand,
                              NULL);
   }
 
@@ -417,12 +429,43 @@ ephy_settings_get_tabs_bar_position (void)
   }
 }
 
+static void
+box_set_halign (GtkWidget       *box,
+                GtkPositionType  type)
+{
+  switch (type) {
+    case GTK_POS_LEFT:
+    case GTK_POS_RIGHT:
+      gtk_widget_set_halign (box, GTK_ALIGN_FILL);
+      break;
+    case GTK_POS_TOP:
+    case GTK_POS_BOTTOM:
+      gtk_widget_set_halign (box, GTK_ALIGN_CENTER);
+      break;
+    default:
+      break;
+  }
+}
+
 static void
 position_changed_cb (GSettings    *settings,
                      char         *key,
                      EphyNotebook *nb)
 {
-  gtk_notebook_set_tab_pos (GTK_NOTEBOOK (nb), ephy_settings_get_tabs_bar_position ());
+  GtkPositionType type = ephy_settings_get_tabs_bar_position ();
+  gint pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (nb));
+  gint i;
+
+  /* Update halign of all notebook label widgets (sub-box) */
+  for (i = 0; i < pages; i++) {
+    GtkWidget *child = gtk_notebook_get_nth_page (GTK_NOTEBOOK (nb), i);
+    GtkWidget *label_widget = gtk_notebook_get_tab_label (GTK_NOTEBOOK (nb), child);
+    GtkWidget *box = gtk_container_get_children (GTK_CONTAINER (label_widget))->data;
+
+    box_set_halign (box, type);
+  }
+
+  gtk_notebook_set_tab_pos (GTK_NOTEBOOK (nb), type);
 }
 
 static void
@@ -433,6 +476,39 @@ show_tabs_changed_cb (GSettings    *settings,
   update_tabs_visibility (nb, FALSE);
 }
 
+static int
+get_last_pinned_tab_pos (EphyNotebook *notebook)
+{
+  int pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook));
+  int found = -1;
+
+  for (int i = 0; i < pages; i++) {
+    GtkWidget *child = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), i);
+
+    if (ephy_notebook_tab_is_pinned (notebook, EPHY_EMBED (child)))
+      found = i;
+  }
+
+  return found;
+}
+
+static void
+page_reordered_cb (GtkNotebook *notebook,
+                   GtkWidget   *child,
+                   guint        page_num,
+                   gpointer     user_data)
+{
+  int last_pinned_tab_pos = get_last_pinned_tab_pos (EPHY_NOTEBOOK (notebook));
+  gboolean is_current_tab_pinned = ephy_notebook_tab_is_pinned (EPHY_NOTEBOOK (notebook), EPHY_EMBED (child));
+
+  /* Ensure that pinned tabs will always stay at the beginning of tab bar */
+  if (last_pinned_tab_pos != -1 && page_num <= (guint)last_pinned_tab_pos) {
+    gint new_pos = is_current_tab_pinned ? last_pinned_tab_pos : last_pinned_tab_pos + 1;
+
+    gtk_notebook_reorder_child (notebook, child, new_pos);
+  }
+}
+
 static void
 ephy_notebook_init (EphyNotebook *notebook)
 {
@@ -452,6 +528,8 @@ ephy_notebook_init (EphyNotebook *notebook)
   g_signal_connect_after (notebook, "switch-page",
                           G_CALLBACK (ephy_notebook_switch_page_cb),
                           NULL);
+  g_signal_connect (notebook, "page-reordered",
+                    G_CALLBACK (page_reordered_cb), NULL);
 
   /* Set up drag-and-drop target */
   g_signal_connect (notebook, "drag-data-received",
@@ -471,6 +549,8 @@ ephy_notebook_init (EphyNotebook *notebook)
   g_signal_connect (EPHY_SETTINGS_UI,
                     "changed::" EPHY_PREFS_UI_TABS_BAR_VISIBILITY_POLICY,
                     G_CALLBACK (show_tabs_changed_cb), notebook);
+
+  gtk_style_context_add_class (gtk_widget_get_style_context (widget), "main-notebook");
 }
 
 static void
@@ -529,7 +609,6 @@ get_nth_tab_label_text (GtkNotebook *notebook,
 {
   GtkWidget *page;
   GtkWidget *tab_label;
-  GtkWidget *label;
 
   g_assert (n >= 0);
 
@@ -537,12 +616,9 @@ get_nth_tab_label_text (GtkNotebook *notebook,
   g_assert (page != NULL);
 
   tab_label = gtk_notebook_get_tab_label (notebook, page);
-  g_assert (GTK_IS_BOX (tab_label));
+  g_assert (EPHY_IS_TAB_LABEL (tab_label));
 
-  label = g_object_get_data (G_OBJECT (tab_label), "label");
-  g_assert (GTK_IS_LABEL (label));
-
-  return gtk_label_get_text (GTK_LABEL (label));
+  return ephy_tab_label_get_text (tab_label);
 }
 
 static char *
@@ -604,69 +680,32 @@ ephy_notebook_rebuild_tab_menu (EphyNotebook *notebook)
 }
 
 static void
-sync_load_status (EphyWebView *view, GParamSpec *pspec, GtkWidget *proxy)
+title_changed_cb (EphyEmbed    *embed,
+                  GParamSpec   *pspec,
+                  EphyNotebook *notebook)
 {
-  GtkWidget *spinner, *icon;
-  EphyEmbed *embed;
+  GtkWidget *tab_label = NULL;
 
-  spinner = GTK_WIDGET (g_object_get_data (G_OBJECT (proxy), "spinner"));
-  icon = GTK_WIDGET (g_object_get_data (G_OBJECT (proxy), "icon"));
-  g_assert (spinner != NULL && icon != NULL);
+  if (ephy_embed_has_load_pending (embed))
+    return;
 
-  embed = EPHY_GET_EMBED_FROM_EPHY_WEB_VIEW (view);
-  if (ephy_web_view_is_loading (view) && !ephy_embed_has_load_pending (embed)) {
-    gtk_widget_hide (icon);
-    gtk_widget_show (spinner);
-    gtk_spinner_start (GTK_SPINNER (spinner));
-  } else {
-    gtk_spinner_stop (GTK_SPINNER (spinner));
-    gtk_widget_hide (spinner);
-    gtk_widget_show (icon);
+  tab_label = gtk_notebook_get_tab_label (GTK_NOTEBOOK (notebook), GTK_WIDGET (embed));
+  g_assert (tab_label);
+
+  ephy_notebook_rebuild_tab_menu (notebook);
+
+  if (ephy_tab_label_is_pinned (tab_label)) {
+    int current_page = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
+    int page_num = gtk_notebook_page_num (GTK_NOTEBOOK (notebook), GTK_WIDGET (embed));
+
+    if (current_page != page_num)
+      ephy_tab_label_set_needs_attention (tab_label, TRUE);
   }
 }
 
 static void
-load_changed_cb (EphyWebView *view, WebKitLoadEvent load_event, GtkWidget *proxy)
-{
-  sync_load_status (view, NULL, proxy);
-}
-
-static void
-sync_icon (EphyWebView *view,
-           GParamSpec  *pspec,
-           GtkImage    *icon)
-{
-  gtk_image_set_from_pixbuf (icon, ephy_web_view_get_icon (view));
-}
-
-static void
-sync_label (EphyEmbed *embed, GParamSpec *pspec, GtkWidget *label)
-{
-  const char *title;
-
-  title = ephy_embed_get_title (embed);
-  gtk_label_set_text (GTK_LABEL (label), title);
-  gtk_widget_set_tooltip_text (label, title);
-}
-
-static void
-rebuild_tab_menu_cb (EphyEmbed    *embed,
-                     GParamSpec   *pspec,
-                     EphyNotebook *notebook)
-{
-  ephy_notebook_rebuild_tab_menu (notebook);
-}
-
-static void
-sync_is_playing_audio (WebKitWebView *view,
-                       GParamSpec    *pspec,
-                       GtkWidget     *speaker_icon)
-{
-  gtk_widget_set_visible (speaker_icon, webkit_web_view_is_playing_audio (view));
-}
-
-static void
-close_button_clicked_cb (GtkWidget *widget, GtkWidget *tab)
+close_button_clicked_cb (GtkWidget *widget,
+                         GtkWidget *tab)
 {
   GtkWidget *notebook;
 
@@ -674,132 +713,37 @@ close_button_clicked_cb (GtkWidget *widget, GtkWidget *tab)
   g_signal_emit (notebook, signals[TAB_CLOSE_REQUEST], 0, tab);
 }
 
-static void
-tab_label_style_set_cb (GtkWidget *hbox,
-                        GtkStyle  *previous_style,
-                        gpointer   user_data)
-{
-  PangoFontMetrics *metrics;
-  PangoContext *context;
-  GtkStyleContext *style;
-  PangoFontDescription *font_desc;
-  GtkWidget *button;
-  int char_width, h, w;
-
-  context = gtk_widget_get_pango_context (hbox);
-  style = gtk_widget_get_style_context (hbox);
-  gtk_style_context_get (style, gtk_style_context_get_state (style),
-                         "font", &font_desc, NULL);
-  metrics = pango_context_get_metrics (context,
-                                       font_desc,
-                                       pango_context_get_language (context));
-  pango_font_description_free (font_desc);
-  char_width = pango_font_metrics_get_approximate_digit_width (metrics);
-  pango_font_metrics_unref (metrics);
-
-  gtk_icon_size_lookup (GTK_ICON_SIZE_MENU, &w, &h);
-
-  gtk_widget_set_size_request (hbox, TAB_WIDTH_N_CHARS * PANGO_PIXELS (char_width) + 2 * w, -1);
-
-  button = g_object_get_data (G_OBJECT (hbox), "close-button");
-  gtk_widget_set_size_request (button, w + 2, h + 2);
-}
-
 static GtkWidget *
-build_tab_label (EphyNotebook *nb, EphyEmbed *embed)
+build_tab_label (EphyNotebook *nb,
+                 EphyEmbed    *embed)
 {
-  GtkWidget *hbox, *label, *close_button, *image, *spinner, *icon, *speaker_icon;
-  GtkWidget *box;
+  GtkWidget *tab_label;
   EphyWebView *view;
 
-  box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 4);
-  gtk_widget_show (box);
-
-  /* set hbox spacing and label padding (see below) so that there's an
-   * equal amount of space around the label */
-  hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
-  gtk_widget_show (hbox);
-  gtk_widget_set_halign (hbox, GTK_ALIGN_CENTER);
-  gtk_box_pack_start (GTK_BOX (box), hbox, TRUE, TRUE, 0);
-
-  /* setup load feedback */
-  spinner = gtk_spinner_new ();
-  gtk_box_pack_start (GTK_BOX (hbox), spinner, FALSE, FALSE, 0);
-
-  /* setup site icon, empty by default */
-  icon = gtk_image_new ();
-  gtk_box_pack_start (GTK_BOX (hbox), icon, FALSE, FALSE, 0);
-  /* don't show the icon */
-
-  /* setup label */
-  label = gtk_label_new (NULL);
-  gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
-  gtk_label_set_single_line_mode (GTK_LABEL (label), TRUE);
-  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
-  gtk_widget_show (label);
-
-  /* setup speaker icon */
-  speaker_icon = gtk_image_new_from_icon_name ("audio-volume-high-symbolic",
-                                               GTK_ICON_SIZE_MENU);
-  gtk_box_pack_start (GTK_BOX (hbox), speaker_icon, FALSE, FALSE, 0);
-
-  /* setup close button */
-  close_button = gtk_button_new ();
-  gtk_button_set_relief (GTK_BUTTON (close_button),
-                         GTK_RELIEF_NONE);
-  /* don't allow focus on the close button */
-  gtk_widget_set_focus_on_click (close_button, FALSE);
-
-  gtk_widget_set_name (close_button, "ephy-tab-close-button");
-
-  image = gtk_image_new_from_icon_name ("window-close-symbolic",
-                                        GTK_ICON_SIZE_MENU);
-  gtk_widget_set_tooltip_text (close_button, _("Close tab"));
-  g_signal_connect (close_button, "clicked",
-                    G_CALLBACK (close_button_clicked_cb), embed);
-
-  gtk_container_add (GTK_CONTAINER (close_button), image);
-  gtk_widget_show (image);
-
-  gtk_box_pack_start (GTK_BOX (box), close_button, FALSE, FALSE, 0);
-  gtk_widget_show (close_button);
-
-  /* Set minimal size */
-  g_signal_connect (box, "style-set",
-                    G_CALLBACK (tab_label_style_set_cb), NULL);
+  tab_label = ephy_tab_label_new ();
+  g_signal_connect (tab_label, "close-clicked", G_CALLBACK (close_button_clicked_cb), embed);
 
   /* Set up drag-and-drop target */
-  g_signal_connect (box, "drag-data-received",
+  g_signal_connect (tab_label, "drag-data-received",
                     G_CALLBACK (notebook_drag_data_received_cb), embed);
-  gtk_drag_dest_set (box, GTK_DEST_DEFAULT_ALL,
+  gtk_drag_dest_set (tab_label, GTK_DEST_DEFAULT_ALL,
                      url_drag_types, G_N_ELEMENTS (url_drag_types),
                      GDK_ACTION_MOVE | GDK_ACTION_COPY);
-  gtk_drag_dest_add_text_targets (box);
-
-  g_object_set_data (G_OBJECT (box), "label", label);
-  g_object_set_data (G_OBJECT (box), "spinner", spinner);
-  g_object_set_data (G_OBJECT (box), "icon", icon);
-  g_object_set_data (G_OBJECT (box), "close-button", close_button);
-  g_object_set_data (G_OBJECT (box), "speaker-icon", speaker_icon);
+  gtk_drag_dest_add_text_targets (tab_label);
 
   /* Hook the label up to the tab properties */
   view = ephy_embed_get_web_view (embed);
-  sync_icon (view, NULL, GTK_IMAGE (icon));
-  sync_label (embed, NULL, label);
-  sync_load_status (view, NULL, box);
-  sync_is_playing_audio (WEBKIT_WEB_VIEW (view), NULL, speaker_icon);
 
-  g_signal_connect_object (view, "notify::icon",
-                           G_CALLBACK (sync_icon), icon, 0);
   g_signal_connect_object (embed, "notify::title",
-                           G_CALLBACK (sync_label), label, 0);
-  g_signal_connect_object (embed, "notify::title",
-                           G_CALLBACK (rebuild_tab_menu_cb), nb, 0);
-  g_signal_connect_object (view, "load-changed",
-                           G_CALLBACK (load_changed_cb), box, 0);
-  g_signal_connect_object (view, "notify::is-playing-audio",
-                           G_CALLBACK (sync_is_playing_audio), speaker_icon, 0);
-  return box;
+                           G_CALLBACK (title_changed_cb), nb, 0);
+
+  g_object_bind_property (view, "title", tab_label, "label-text", G_BINDING_DEFAULT);
+  g_object_bind_property (view, "display-address", tab_label, "label-uri", G_BINDING_DEFAULT);
+  g_object_bind_property (view, "icon", tab_label, "icon-buf", G_BINDING_DEFAULT);
+  g_object_bind_property (view, "is-loading", tab_label, "spinning", G_BINDING_DEFAULT);
+  g_object_bind_property (view, "is-playing-audio", tab_label, "audio", G_BINDING_DEFAULT);
+
+  return tab_label;
 }
 
 void
@@ -913,9 +857,7 @@ ephy_notebook_remove (GtkContainer *container,
 {
   GtkNotebook *gnotebook = GTK_NOTEBOOK (container);
   EphyNotebook *notebook = EPHY_NOTEBOOK (container);
-  GtkWidget *tab_label, *tab_label_label, *tab_label_icon, *tab_label_speaker_icon;
   int position, curr;
-  EphyWebView *view;
 
   if (!EPHY_IS_EMBED (tab_widget))
     return;
@@ -926,31 +868,13 @@ ephy_notebook_remove (GtkContainer *container,
   position = gtk_notebook_page_num (gnotebook, tab_widget);
   curr = gtk_notebook_get_current_page (gnotebook);
 
-  if (position == curr) {
+  if (position == curr)
     smart_tab_switching_on_closure (notebook, tab_widget);
-  }
-
-  /* Prepare tab label for destruction */
-  tab_label = gtk_notebook_get_tab_label (gnotebook, tab_widget);
-  tab_label_icon = g_object_get_data (G_OBJECT (tab_label), "icon");
-  tab_label_label = g_object_get_data (G_OBJECT (tab_label), "label");
-  tab_label_speaker_icon = g_object_get_data (G_OBJECT (tab_label), "speaker-icon");
-
-  view = ephy_embed_get_web_view (EPHY_EMBED (tab_widget));
-
-  g_signal_handlers_disconnect_by_func
-    (view, G_CALLBACK (sync_icon), tab_label_icon);
-  g_signal_handlers_disconnect_by_func
-    (tab_widget, G_CALLBACK (sync_label), tab_label_label);
-  g_signal_handlers_disconnect_by_func
-    (tab_widget, G_CALLBACK (sync_label), notebook);
-  g_signal_handlers_disconnect_by_func
-    (view, G_CALLBACK (sync_load_status), tab_label);
-  g_signal_handlers_disconnect_by_func
-    (view, G_CALLBACK (sync_is_playing_audio), tab_label_speaker_icon);
 
   GTK_CONTAINER_CLASS (ephy_notebook_parent_class)->remove (container, tab_widget);
 
+  g_signal_handlers_disconnect_by_data (tab_widget, notebook);
+
   update_tabs_visibility (notebook, FALSE);
 }
 
@@ -1066,3 +990,35 @@ ephy_notebook_set_adaptive_mode (EphyNotebook     *notebook,
   notebook->adaptive_mode = adaptive_mode;
   update_tabs_visibility (notebook, FALSE);
 }
+
+void
+ephy_notebook_tab_set_pinned (EphyNotebook *notebook,
+                              GtkWidget    *embed,
+                              gboolean      is_pinned)
+{
+  GtkWidget *tab_label;
+  gboolean expanded;
+
+  gtk_notebook_set_tab_reorderable (GTK_NOTEBOOK (notebook), embed, !is_pinned);
+  tab_label = gtk_notebook_get_tab_label (GTK_NOTEBOOK (notebook), embed);
+  ephy_tab_label_set_pinned (tab_label, is_pinned);
+
+  if (is_pinned) {
+    gtk_notebook_reorder_child (GTK_NOTEBOOK (notebook), embed, 0);
+    expanded = FALSE;
+  } else {
+    expanded = g_settings_get_boolean (EPHY_SETTINGS_UI, EPHY_PREFS_UI_EXPAND_TABS_BAR);
+    gtk_notebook_reorder_child (GTK_NOTEBOOK (notebook), embed, -1);
+  }
+
+  gtk_container_child_set (GTK_CONTAINER (notebook), embed, "tab-expand", expanded, NULL);
+}
+
+gboolean
+ephy_notebook_tab_is_pinned (EphyNotebook *notebook,
+                             EphyEmbed    *embed)
+{
+  GtkWidget *tab_label = gtk_notebook_get_tab_label (GTK_NOTEBOOK (notebook), GTK_WIDGET (embed));
+
+  return ephy_tab_label_is_pinned (tab_label);
+}
